// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file MultiGridSolver.h */

#ifndef AMDIS_MULTIGRIDSOLVER_H
#define AMDIS_MULTIGRIDSOLVER_H

#include "MultiGridSolverBase.h"
#include <vector>
#include "Global.h"
#include "SmootherBase.h"
#include <set>
#include <map>
#include "MatrixVector.h"
#include "SystemVector.h"

namespace AMDiS {

  class DOFMatrix;
  template<typename T> class DOFVector;
  template<typename T> class SparseVector;
  class LevelAdmin;
  class FiniteElemSpace;
  class DOFAdmin;
  class RCNeighbourList;
  class InterpolRestrictMatrix;

  /** \brief
   * Implementation of the multigrid solver base class for scalar problems.
   */
  class MultiGridSolverScal : public MultiGridSolverBase<DOFMatrix, DOFVector<double> >
  {
  protected:
    /** \brief
     * Class to assemble all data of one multigrid level.
     */
    class LevelData
    {
    public:
      LevelData() 
	: matrix_(NULL),
	  solution_(NULL),
	  oldSolution_(NULL),
	  rhs_(NULL)
      {};

      /** \brief
       * The level DOFMatrix.
       */
      DOFMatrix                *matrix_;

      /** \brief
       * The solution stored in sparse format.
       */
      SparseVector<double>     *solution_;

      /** \brief
       * The old solution stored in sparse format.
       */
      SparseVector<double>     *oldSolution_;

      /** \brief
       * The right hand side of the level stored in sparse format.
       */
      SparseVector<double>     *rhs_;

      /** \brief
       * Set of DOFs of all elements of the level.
       */
      ::std::set<DegreeOfFreedom> levelDOFs_;

      /** \brief
       * Set of DOFs of all elements of the level, which are also 
       * included in finer levels.
       */
      ::std::set<DegreeOfFreedom> fineDOFs_;

      /** \brief
       * Set of DOFs of all elements of the level, which are not
       * included in finer levels.
       */
      ::std::set<DegreeOfFreedom> coarseDOFs_;

      /** \brief
       */
      ::std::set<DegreeOfFreedom> nearestNeighbourDOFs_;

      /** \brief
       * List of all elements in the mesh on the current level.
       */
      ::std::vector<Element*>     elements_;

      /** \brief
       * This array is only used if the elements are of dim=3. Than there
       * are different element types possible. All different element types
       * of the mesh are stored here.
       */
      ::std::vector<int>          elementTypes_;
    };

  public:
    /** \brief
     * Constructor of the scalar multigrid solver.
     */
    MultiGridSolverScal(const ::std::string &name);

    /** \brief
     * Computes the residual and checks if its norm is less than the tolerance.
     */
    bool toleranceReached(double tol);

    /** \brief
     * Traverses all elements and searchs for the highest level of an 
     * elements DOF. The result is stored in variable maxLevel_.
     */
    void initMaxLevel();

    /** \brief
     * Initialize the multigrid solver.
     */
    void initMultiGrid(bool initMG);

    /** \brief
     * Frees all allocated memory.
     */
    void exitMultiGrid();

    /** \brief
     * Solve the given level exactly.
     * TODO: Implement a real exact solver. At the moment we use the smoother.
     */
    void exactSolve(int level);

    /** \brief
     * Call the smoother on the given level. steps is the maximum number of 
     * smoothing steps.
     */
    void smooth(int level, int steps);

    /** \brief
     * Restricts the residual and the solution from a fine multigrid level to 
     * a coarser multigrid level.
     */
    void restrict(int level, int coarseLevel);

    /** \brief
     * Prolongates the residual from a coarse multigrid level to a finer
     * multigrid level.
     */
    void prolongate(int level, int coarseLevel);

    /** \brief
     * Computes the residual of a given leve. CoarseLevel parameter will be
     * not considered.
     */
    void computeResidual(int level, int coarseLevel);

    /** \brief
     * Computes the rhs of the coarse level. Level parameter will be not 
     * considered.
     */
    void computeRHS(int level, int coarseLevel);

    /** \brief
     * Computes the correction of the coarse level. Level parameter will be
     * not considered.
     */
    void computeCorrection(int level, int coarseLevel);

    /** \brief
     * Add the correction, which has been computed on the coarse grid, to the
     * solution of the fine grid.
     */
    void fixSolution(int level, int coarseLevel);

    /** \brief
     * Returns for a given level the next coarser multigrid level.
     */
    int coarserLevel(int level) {
      FUNCNAME("MultiGridSolverScal::coarserLevel()");
      TEST_EXIT(level > minLevel_)("invalid level\n");
      do {
	--level;
      } while(!isMGLevel(level));
      return level;
    };

    /** \brief
     * Returns true, if the level is a multigrid level, otherwise false.
     */
    inline bool isMGLevel(int level) {
      return isMGLevel_[level];
    };

    /** \brief
     * Returns for a given level the next multigrid level.
     */
    inline int mgLevel(int level) {
      while (!isMGLevel(level)) {
	level++;
      }
      return level;    
    };

    /** \brief
     * Initialize the full multigrid cycle.
     */
    void initFullMultiGrid();

    /** \brief
     *
     */
    void fmgInterpol(int level);

  public:
    /** \brief
     * Data of all multigrid levels.
     */
    ::std::vector<LevelData> levels_;

    /** \brief
     * Residual stored in sparse format.
     */
    SparseVector<double> *sparseResidual_;

    /** \brief
     * Residual stored in dense format.
     */
    DOFVector<double> *denseResidual_;

    /** \brief
     * 2-Norm of the last residual.
     */
    double lastRes_;

    /** \brief
     * Defines for all mesh levels, if the level is a multigrid level.
     * isMGLevel[0] is the coarsest grid level.
     */
    ::std::vector<bool> isMGLevel_;

    /** \brief
     * Counts the number of multigrid iterations.
     */
    int iteration_;

    /** \brief
     * Number of smoothing steps on the coarsest level.
     */
    int coarseSmoothingSteps_;
  
    /** \brief
     * Number of mesh levels that at leatst are skipped between two multigrid 
     * levels.
     */
    int minLevelGap_;

    /** \brief
     * Maximum number of allowed multigrid levels.
     */
    int maxMGLevels_;

    /** \brief
     * Smoother
     */
    SmootherBase<DOFMatrix,
		 SparseVector<double>,
		 ::std::set<DegreeOfFreedom> > *smoother_;

    /** \brief
     * Interpolation matrix for interpolating the residual between two meshes.
     */
    InterpolRestrictMatrix *interpolMatrix_;

    /** \brief
     * Restriction matrix for restricting the residual between two meshes.
     */
    InterpolRestrictMatrix *restrictMatrix_;

    /** \brief
     *
     */
    int galerkin_;
  };


  // ============================================================================
  // ============================================================================
  // ============================================================================
  // ============================================================================


  /** \brief
   * Implementation of the multigrid solver base class for scalar problems.
   */
  class MultiGridSolverVec : public MultiGridSolverBase<Matrix<DOFMatrix*>, 
							SystemVector>
  {
  public:
    /** \brief
     * Multigrid data that depends on the fe space.
     */
    class FESpaceData
    {
    public:
      MEMORY_MANAGED(FESpaceData);

      /** \brief
       * Set of DOFs of all elements of the level.
       */
      ::std::set<DegreeOfFreedom> levelDOFs_;

      /** \brief
       * Set of DOFs of all elements of the level, which are also 
       * included in finer levels.
       */
      ::std::set<DegreeOfFreedom> fineDOFs_;

      /** \brief
       * Set of DOFs of all elements of the level, which are not
       * included in finer levels.
       */
      ::std::set<DegreeOfFreedom> coarseDOFs_;

      /** \brief
       */
      ::std::set<DegreeOfFreedom> nearestNeighbourDOFs_;
    };

    /** \brief
     * Multigrid data that depend on the mesh.
     */
    class MeshData
    {
    public:
      MEMORY_MANAGED(MeshData);

      MeshData() : maxLevel(0) {};

      int maxLevel;
    };

    /** \brief
     * Data for one component within one level.
     */
    class ComponentData
    {
    public:
      MEMORY_MANAGED(ComponentData);

      ComponentData()
	: feSpaceData_(NULL),
	  meshData_(NULL)
      {};

      FESpaceData *feSpaceData_;
      MeshData    *meshData_;
    };

    /** \brief
     * Data for one multigrid level.
     */
    class LevelData
    {
    public:
      MEMORY_MANAGED(LevelData);

      LevelData()
	: matrix_(NULL),
	  oldSolution_(NULL),
	  solution_(NULL),
	  rhs_(NULL)
      {};
    
      /** \brief
       * The DOF matrices.
       */
      Matrix<DOFMatrix*>                *matrix_;

      /** \brief
       * Vector of old solutions, each stored in sparse format.
       */
      Vector<SparseVector<double>*>     *oldSolution_;

      /** \brief
       * Vector of solutions, each stored in sparse format.
       */
      Vector<SparseVector<double>*>     *solution_;

      /** \brief
       * Vector of rhs', each stored in sparse format.
       */
      Vector<SparseVector<double>*>     *rhs_;

      /** \brief
       * Mesh and FE space dependend data for each component.
       */
      ::std::vector<ComponentData>         components_;

      /** \brief
       * Stores for each component all DOFs at this level.
       */
      Vector< ::std::set<DegreeOfFreedom>*> levelDOFs_;

      /** \brief
       * List of all elements in the mesh on the current level.
       */
      ::std::vector<Element*>     elements_;

      /** \brief
       * This array is only used if the elements are of dim=3. Than there
       * are different element types possible. All different element types
       * of the mesh are stored here.
       */
      ::std::vector<int>          elementTypes_;
    };

  public:
    /** \brief
     * Constructor of the vector multigrid solver.
     */
    MultiGridSolverVec(const ::std::string &name);

    /** \brief
     * Computes the residual and checks if its norm is less than the tolerance.
     */
    bool toleranceReached(double tol);

    /** \brief
     * Traverses all elements and searchs for the highest level of an 
     * elements DOF. The result is stored in variable maxLevel_.
     */
    void initMaxLevel();

    /** \brief
     * Initialize the multigrid solver.
     */
    void initMultiGrid(bool initMG);

    /** \brief
     * Frees all allocated memory.
     */
    void exitMultiGrid();

    /** \brief
     * Solve the given level exactly.
     * TODO: Implement a real exact solver. At the moment we use the smoother.
     */
    void exactSolve(int level);

    /** \brief
     * Call the smoother on the given level. steps is the maximum number of 
     * smoothing steps.
     */
    void smooth(int level, int steps);

    /** \brief
     * Restricts the residual and the solution from a fine multigrid level to 
     * a coarser multigrid level.
     */
    void restrict(int level, int coarseLevel);

    /** \brief
     * Prolongates the residual from a coarse multigrid level to a finer
     * multigrid level.
     */
    void prolongate(int level, int coarseLevel);

    /** \brief
     * Computes the residual of a given leve. CoarseLevel parameter will be
     * not considered.
     */
    void computeResidual(int level, int coarseLevel);

    /** \brief
     * Computes the rhs of the coarse level. Level parameter will be not 
     * considered.
     */
    void computeRHS(int level, int coarseLevel);

    /** \brief
     * Computes the correction of the coarse level. Level parameter will be
     * not considered.
     */
    void computeCorrection(int level, int coarseLevel);

    /** \brief
     * Add the correction, which has been computed on the coarse grid, to the
     * solution of the fine grid.
     */
    void fixSolution(int level, int coarseLevel);

    /** \brief
     * Returns for a given level the next coarser multigrid level.
     */
    int coarserLevel(int level) {
      FUNCNAME("MultiGridSolverVec::coarserLevel()");
      TEST_EXIT(level > minLevel_)("invalid level\n");
      do {
	--level;
      } while(!isMGLevel(level));
      return level;
    };

    /** \brief
     * Returns true if the level is a multigrid level, otherwise false.
     */
    inline bool isMGLevel(int level) {
      return isMGLevel_[level];
    };

    void initFullMultiGrid();

    void fmgInterpol(int level);

    inline int mgLevel(int level) {
      while(!isMGLevel_[level]) {
	++level;
      }
      return level;    
    };

  protected:
    /** \brief
     * Defines for all mesh levels, if the level is a multigrid level.
     * isMGLevel[0] is the coarsest grid level.
     */
    ::std::vector<bool> isMGLevel_;

    /** \brief
     * Data of all multigrid levels.
     */
    ::std::vector<LevelData> levels_;

    /** \brief
     * Sparse vectors containing the residuals.
     */
    Vector<SparseVector<double>*> *sparseResidual_;

    /** \brief
     * Dense vectors containing the residuals.
     */
    SystemVector *denseResidual_;

    /** \brief
     * 2-Norm of the last residual.
     */
    double lastRes_;

    /** \brief
     * Counts the number of multigrid iterations.
     */
    int iteration_;

    /** \brief
     * Number of smoothing steps on the coarsest level.
     */
    int coarseSmoothingSteps_;
  
    /** \brief
     * Is the number of mesh levels that at least are skipped between two
     * multigrid levels.
     */
    int minLevelGap_;

    /** \brief
     * Maximum number of multigrid levels.
     */
    int maxMGLevels_;

    /** \brief
     * Number of components in the vector problem.
     */
    int numComponents_;

    /** \brief
     * Smoother
     */
    SmootherBase<Matrix<DOFMatrix*>, 
		 Vector<SparseVector<double>*>,
		 Vector< ::std::set<DegreeOfFreedom>*> > *smoother_;

    /** \brief
     * Vector of interpolation matrices for interpolating the residual between 
     * two meshes.
     */
    ::std::vector<InterpolRestrictMatrix*> interpolMatrix_;

    /** \brief
     * Vector of restriction matrices for restricting the residual between 
     * two meshes.
     */
    ::std::vector<InterpolRestrictMatrix*> restrictMatrix_;

    /** \brief
     * Diagonal entries in the system matrix.
     */
    SystemVector *diagEntries_;

    /** \brief
     *
     */
    bool boxSmoothing_;

    /** \brief
     *
     */
    Vector< ::std::set<DegreeOfFreedom>*> boxDOFs_;

    /** \brief
     *
     */
    int galerkin_;
  };

}

#endif
