#include "Global.h"
#include "Parameters.h"

namespace AMDiS {

  template<typename MatrixType, typename VectorType>
  MultiGridSolverBase<MatrixType, VectorType>::MultiGridSolverBase(const ::std::string &name)
    : name_(name),
      tolerance_(1e-8),
      maxIterations_(100),
      minLevel_(0),
      maxLevel_(-1),
      numCycles_(1),
      preSmoothingSteps_(1),
      postSmoothingSteps_(1),
      fmg_(0)
  {
    GET_PARAMETER(0, name_ + "->tolerance", "%e", &tolerance_);
    GET_PARAMETER(0, name_ + "->max iteration", "%d", &maxIterations_);
    GET_PARAMETER(0, name_ + "->mg cycle type", "%d", &numCycles_);
    GET_PARAMETER(0, name_ + "->pre smoothing steps", "%d", &preSmoothingSteps_);
    GET_PARAMETER(0, name_ + "->post smoothing steps", "%d", &postSmoothingSteps_);
    GET_PARAMETER(0, name_ + "->min level", "%d", &minLevel_);
    GET_PARAMETER(0, name_ + "->full mg", "%d", &fmg_);
  }

  template<typename MatrixType, typename VectorType>
  int MultiGridSolverBase<MatrixType, VectorType>::solve(MatrixType &a, 
							 VectorType &x,
							 VectorType &b,
							 bool initMG,
							 bool exitMG)
  {
    int i = 0;
    systemMatrix_ = &a;
    solution_ = &x;
    rhs_ = &b;

    initMaxLevel();
    if (minLevel_ > maxLevel_) 
      minLevel_ = maxLevel_;

    initMultiGrid(initMG);

    if (fmg_) {
      initFullMultiGrid();
      for (int i = 1; i < maxLevel_ + 1; i++) {
	fmgInterpol(i);
	multiGridIteration(i);
      }
    }

    for (int i = 0; i < maxIterations_; i++) {
      if (toleranceReached(tolerance_)) 
	break;
      multiGridIteration(maxLevel_);
    }

    if (exitMG) {
      exitMultiGrid();
    }

    return i + 1;
  }

  template<typename MatrixType, typename VectorType>
  void MultiGridSolverBase<MatrixType, VectorType>::multiGridIteration(int level)
  {
    if (level <= minLevel_) {
      // If we are on the coarsest possible grid, use exact solver.
      exactSolve(minLevel_);
    } else {
      // Get the next coarser level.
      int coarseLevel = coarserLevel(level);

      // Smooth the solution on the fine grid.
      smooth(level, preSmoothingSteps_);

      // Compute the residual on the fine grid.
      computeResidual(level, coarseLevel);

      // Restrict the residual to the coarser grid.
      restrict(level, coarseLevel);

      // Compute defect equation on the coarser grid.
      computeRHS(level, coarseLevel);
      
      // Apply multigrid on the coarser grid.
      for (int i = 0; i < numCycles_; i++) {
	multiGridIteration(coarseLevel);
      }

      // Compute correction of the solution.
      computeCorrection(level, coarseLevel);

      // Prolongate the solution to the finer grid.
      prolongate(level, coarseLevel);

      // Add the correction, computed on the coarse grid, to the solution
      // on the fine grid.
      fixSolution(level, coarseLevel);

      // Apply smoothing on the solution on the fine grid.
      smooth(level, postSmoothingSteps_);
    }
  }

}
