#include "AMDiS.h"
#include "GNUPlotWriter.h"

using namespace std;
using namespace AMDiS;

// ===========================================================================
// ===== function definitions ================================================
// ===========================================================================

/** \brief
 * Dirichlet boundary function
 */
class G : public AbstractFunction<double, WorldVector<double> >
{
public:
  MEMORY_MANAGED(G);

  /** \brief
   * Implementation of AbstractFunction::operator().
   */
  const double& operator()(const WorldVector<double>& x) const
  {
    static double result;
    result = exp(-10.0*(x*x));
    return result;  
  };
};

/** \brief
 * RHS function
 */
class F : public AbstractFunction<double, WorldVector<double> >
{
public:
  MEMORY_MANAGED(F);

  F(int degree) : AbstractFunction<double, WorldVector<double> >(degree) {};

  /** \brief
   * Implementation of AbstractFunction::operator().
   */
  const double& operator()(const WorldVector<double>& x) const {
    static double result = 0.0;
    int dim = x.getSize();
    double r2 = (x*x);
    double ux = exp(-10.0*r2);
    result = -(400.0*r2 - 20.0*dim)*ux;
    return result;
  };
};

// ===========================================================================
// ===== main program ========================================================
// ===========================================================================

int main(int argc, char* argv[])
{
  FUNCNAME("main");

  // ===== check for init file =====
  TEST_EXIT(argc == 2)("usage: vecmultigrid initfile\n");

  // ===== init parameters =====
  Parameters::init(false, argv[1]);

  // ===== create and init the scalar problem ===== 
  ProblemVec vecmultigrid("vecmultigrid");
  vecmultigrid.initialize(INIT_ALL);

  // === create adapt info ===
  AdaptInfo *adaptInfo = NEW AdaptInfo("vecmultigrid->adapt",
				       vecmultigrid.getNumComponents());

  // === create adapt ===
  AdaptStationary *adapt = NEW AdaptStationary("vecmultigrid->adapt",
					       &vecmultigrid,
					       adaptInfo);
  

  // ===== add boundary conditions =====
  vecmultigrid.addDirichletBC(1, 0, NEW G);
  //vecmultigrid.addDirichletBC(1, 1, NEW G);

  // ===== create operators =====
  Operator matrixOperator00(Operator::MATRIX_OPERATOR,
			    vecmultigrid.getFESpace(0),
			    vecmultigrid.getFESpace(0));
  matrixOperator00.addSecondOrderTerm(NEW Laplace_SOT);
  vecmultigrid.addMatrixOperator(&matrixOperator00, 0, 0);

  Operator rhsOperator0(Operator::VECTOR_OPERATOR,
			vecmultigrid.getFESpace(0));

  int degree = vecmultigrid.getFESpace(0)->getBasisFcts()->getDegree();

  rhsOperator0.addZeroOrderTerm(NEW CoordsAtQP_ZOT(NEW F(degree)));

  vecmultigrid.addVectorOperator(&rhsOperator0, 0);

  Operator matrixOperator10(Operator::MATRIX_OPERATOR,
			    vecmultigrid.getFESpace(1),
			    vecmultigrid.getFESpace(0));

  Operator matrixOperator11(Operator::MATRIX_OPERATOR,
			    vecmultigrid.getFESpace(1),
			    vecmultigrid.getFESpace(1));

  matrixOperator10.addZeroOrderTerm(NEW Simple_ZOT);

  vecmultigrid.addMatrixOperator(&matrixOperator10, 1, 0);

  matrixOperator11.addZeroOrderTerm(NEW Simple_ZOT(-1.0));

  vecmultigrid.addMatrixOperator(&matrixOperator11, 1, 1);

  // ===== start adaption loop =====
  adapt->adapt();

  vecmultigrid.writeFiles(adaptInfo, true);

  return 0;
}


